module Sem10 where

import Control.Applicative(liftA2)

-- Монады.
-- Можно сколько угодно рассуждать о том, что же такое монада. Есть теоретико категориальное определение,
-- есть более программистское, которое вы видели на лекции, есть издевательское("монада -- это всего лишь моноид
-- в категории эндофункторов"), но мы не будем этого делать.
-- Стремный пример для затравки. Чему равно значение выражения(да, я решил их вернуть)?
-- [1,2] >>= \n -> ['a', 'b'] >>= \c -> return (n, c)
-- Как записать его в do-нотации?

example1 :: [(Integer, Char)]
example1 = do
    n <- [1,2]
    c <- ['a', 'b']
    return (n, c)

-- Напишите реализации функций filter и replicate через do-нотацию и монаду списка

filter' :: (a -> Bool) -> [a] -> [a]
filter' p xs = do
    x <- xs
    True <- return $ p x
    return x


-- В определении выше паттерматчинг при неудаче вызовет fail
-- В определении ниже -- не вызовет. С большой долей вероятности у вас выстрелит эксепшен о не всех разобранных образцах
filter'' :: (a -> Bool) -> [a] -> [a]
filter'' p xs = xs >>= \x -> return (p x) >>= \True -> return x

-- Как это примерно работает(с учетом того, что я сказал выше)
-- filter'' odd [1,2,3] =>
-- [1,2,3] >>= \x -> return (odd x) >>= \True -> return x =>
-- concat (fmap (\x -> return (odd x) >>= \True -> return x) [1,2,3])
-- concat [return (odd 1) >>= \True -> return 1, return (odd 2) >>= \True -> return 2], [return (odd 3) >>= \True -> return 3] =>
-- concat [(concat (fmap (\True -> return 1) [True]), (concat (fmap (\True -> return 2) [False]), (concat (fmap (\True -> return 3) [True])] =>
-- concat [(concat [[1]]), (HERE COMES THE FAIL), (concat [[3]])] =>
-- concat [[1], [], [3]] => [1,3]

replicate' :: Int -> a -> [a]
replicate' n x = do
    [1..n]
    return x

replicate'' :: Int -> a -> [a]
replicate'' n x = [1..n] >> [x] -- [1..n] >>= const x

-- Что вернет вызов и почему?
-- GHCi> (\bs -> do {b <- bs; when b []; return 42}) [True, False, True]

-- do {b <- [True, False, True]; when b []; return 42} =>
-- [True, False, True] >>= (\b -> do {when b []; return 42}) =>
-- [True, False, True] >>= \b -> when b [] >> return 42 =>
-- [True, False, True] >>= \b -> when b [] >>= \_ -> return 42 =>
-- concat [[] >>= \_ -> return 42, [()] >>= \_ -> return 42, [] >>= \_ return 42] =>
-- concat [[], [42], []] => [42]


-- Лирическое отступление. В полном определении класса типов Traversable есть
-- монадические аналоги traverse и sequenceA. Первый зовется mapM, а второй -- sequence
-- Их определения до ужаса просты, они просто вызывают соответствующих аппликативных родителей

-- Кроме того, в Control.Monad живут монадические обобщения некоторых функций над списками
-- Например zipWithM, replicateM, filterM и foldM

zipWithM :: (Applicative m) => (a -> b -> m c) -> [a] -> [b] -> m [c]
zipWithM f xs ys = sequenceA (zipWith f xs ys)

replicateM :: (Applicative m) => Int -> m a -> m [a]
replicateM cnt0 f =
    loop cnt0
        where
            loop cnt
                | cnt <= 0  = pure []
                | otherwise = liftA2 (:) f (loop (cnt - 1))

filterM :: (Applicative m) => (a -> m Bool) -> [a] -> m [a]
filterM p = foldr (\ x -> liftA2 (\ flg -> if flg then (x:) else id) (p x)) (pure [])

foldM :: (Foldable t, Monad m) => (b -> a -> m b) -> b -> t a -> m b
foldM f z0 xs = foldr f' return xs z0
  where f' x k z = f z x >>= k

-- Выше приведены их определения, стащенные из хугла. В новой версии компилятора у первых трех
-- функций контекст Monad был обобщен до Applicative, а у foldM список обобщился до произвольного
-- Foldable. Давайте посмотрим на примеры использования этих чуваков:

-- Повызывайте в GHCi:
-- zipWithM (\x y -> Just (x + y)) [1,2,3] [4, 5]
-- zipWithM (\x y -> (show x ++ "+" ++ show y ++ ";", x + y)) [1,2,3] [4,5,6]
-- replicateM 2 (ZipList [1,2,3])
-- filterM (Just . odd) [1,2,3]
-- traverse (Just . odd) [1,2,3]
-- filterM (\_ -> [True, False]) [1,2,3]

-- Имея foldM можно, например суммировать числа, контроллируя выход за диапазон

isTiny :: (Ord a, Num a) => a -> Bool
isTiny x = x > (-128) && x < 127

(?+?) :: (Ord a, Num a) => a -> a -> Maybe a
x ?+? y = let res = x + y in if isTiny res then Just res else Nothing

infixl 6 ?+?

-- Повызывайте в GHCi:
-- 100 ?+? 1
-- 100 ?+? 50
-- foldM (?+?) 0 [1..15]
-- foldM (?+?) 0 [1..16]

-- Поговорим теперь, про то, как связаны монады, функторы и аппликативные функторы.
-- Давайте покажем, что имея на руках монаду, можно без особого труда выразить
-- методы Functor и Applicative

-- Выразите fmap через (>>=) и return
fmap' :: Monad m => (a -> b) -> m a -> m b
fmap' f x = x >>= return . f

-- И через do-нотацию
fmap'' :: Monad m => (a -> b) -> m a -> m b
fmap'' f x = do
    v <- x
    return $ f v

-- То же самое с Applicative, напишите (<*>) используя return и (>>=)

(<*>.) :: Monad m => m (a -> b) -> m a -> m b
(<*>.) fs xs = fs >>= \f -> xs >>= return . f

-- И do-нотацию
(.<*>.) :: Monad m => m (a -> b) -> m a -> m b
(.<*>.) fs xs = do
    f <- fs
    x <- xs
    return $ f x

-- Заметим, что выше, вообще говоря, надо бы еще показать, что законы функторов следуют из
-- законов монад, равно как и законы аппликативных функторов. Вместо этого давайте рассмотрим вот такие
-- вот монадические комбинаторы, определенные в Control.Monad

-- Этот зовется Kleisli composition, но все зовут его рыбкой
-- Он позволяет задать композию стрелок Клейсли
-- >=> :: Monad m => (a -> m b) -> (b -> m c) -> (a -> m c)

-- У него есть повернутый брат-близнец
-- <=< :: Monad m => (b -> m c) -> (a -> m b) -> (a -> m c)
-- <=< = flip (>=>)

-- И, наконец:
-- join :: Monad m => m (m a) -> m a

-- Выразите (>=>) через >>=
(>=>.) :: Monad m => (a -> m b) -> (b -> m c) -> (a -> m c)
(>=>.) f g = \x -> f x >>= g

-- Выразите join через (>>=)
join' :: Monad m => m (m a) -> m a
join' x = x >>= id

-- Выразите (>=>) через join и fmap
(.>=>.) :: Monad m => (a -> m b) -> (b -> m c) -> (a -> m c)
(.>=>.) f g = join' . fmap g . f

-- Переведите законы монад из формулировок через (>>=) и return, в формулировку через (>=>) и return
-- Сделайте то же самое, но теперь уже в формулировку через join (и fmap) и return. Для второго полезно
-- помнить закон функтора и "свободные теоремы" для типов return и join

-- Исходные формулировки монадных законов:
-- return x >>= k   = k x
-- m >>= return     = m
-- m >>= u >>= v    = m >>= (\x -> u x >>= v)

-- Первый
-- f >=> g = \x -> f x >>= g
-- return >=> k = k

-- Второй. Пусть m = f x
-- f x >>= return = f x
-- (f >=> return) x = f x
-- f >=> return = f

-- Третий. Преобразований тут много, но они довольно механические
-- f >=> g >=> h = f >=> (g >=> h)

-- Теперь через join, fmap и return
-- Первый
-- m >>= k = (join . fmap k) m
-- return x >>= k = k x
-- join (fmap k (return x)) = k x
-- join . return = id

-- Второй
-- f x >>= return = f x
-- join (fmap return (f x)) = f x
-- join . fmap return = id

-- Третий. Преобразований тут еще больше, и вот тут и оказываются полезными
-- утверждения, приведенные ниже
-- join . join = join . fmap join

-- Законы функторов:
-- fmap id      = id
-- fmap (f . g) = fmap f . fmap g

-- Свободные теоремы для типов return и join:
-- return . f           = fmap f . return
-- join . fmap (fmap f) = fmap f . join


-- Пример забавной монады
-- У нас было вот такое определение лямбда-термов и свободные переменные мы брали просто из строк
data Term = Var String
    | App Term Term
    | Lam String Term


-- Занятно, но его можно обобщить и брать свободные переменные из множества v
-- Тогда для того, что бы абстрагироваться, нам нужен терм, у которого на
-- одну свободную переменную больше. Простейший способ сделтать это -- взять
-- Maybe v. Тогда все свободные переменные в v можно сложить в Just, а новая
-- свободная переменная -- это Nothing

data Term' v = Var' v
    | App' (Term' v) (Term' v)
    | Lam' (Term' (Maybe v))

-- Тогда определение выше -- функтор. fmap просто переименовывает переменные
-- Оно же и монада! Это не очень сложно доказать формально, но мы не будем
-- этого делать. return -- это просто конструктор переменной, а >>= -- это,
-- внезапно, подстановка. Из типа (>>=) :: Term' a -> (a -> Term' b) -> Term' b
-- это не очень очевидно, но если вы вспомните, что стрелки из a в Term' b
-- можно задавать в виде списка пар (x, t), то станет чуточку яснее. Ну
-- действительно, мы по факту, даем подстановке список из переменный и того,
-- что должно быть подставлено на ее место. Монадные законы для этого типа
-- тогда описывают свойства подстановки
